# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import logging
from glob import glob
from nn_meter import list_latency_predictors, load_latency_predictor, model_file_to_graph


def list_latency_predictors_cli():
    preds = list_latency_predictors()
    logging.keyinfo("Supported latency predictors:")
    for p in preds:
        logging.result(f"[Predictor] {p['name']}: version={p['version']}")
    return


def apply_latency_predictor_cli(args):
    """apply latency predictor to predict model latency according to the command line interface arguments
    """

    # specify model type
    if args.tensorflow:
        input_model, model_type, model_suffix = args.tensorflow, "pb", ".pb"
    elif args.onnx:
        input_model, model_type, model_suffix = args.onnx, "onnx", ".onnx"
    elif args.nn_meter_ir:
        input_model, model_type, model_suffix = args.nn_meter_ir, "nnmeter-ir", ".json"
    elif args.torchvision: # torch model name from torchvision model zoo
        input_model_list, model_type = args.torchvision, "torch" 
    else:
        logging.keyinfo('please run "nn-meter predict --help" to see guidance.')
        return
    
    if not args.predictor:
        logging.keyinfo('You must specify a predictor. Use "nn-meter --list-predictors" to see all supporting predictors.')
        return

    # load predictor
    predictor = load_latency_predictor(args.predictor, args.predictor_version)

    # specify model for prediction
    if not args.torchvision: # input of tensorflow, onnx, nnmeter-ir and nni-ir is file name, while input of torchvision is string list
        input_model_list = []
        if os.path.isfile(input_model):
            input_model_list = [input_model]
        elif os.path.isdir(input_model):
            input_model_list = glob(os.path.join(input_model, "**" + model_suffix))
            input_model_list.sort()
            logging.info(f'Found {len(input_model_list)} model in {input_model}. Start prediction ...')
        else:
            logging.error(f'Cannot find any model satisfying the arguments.')

    # predict latency
    result = {}
    for model in input_model_list:
        latency = predictor.predict(model, model_type) # in unit of ms
        result[os.path.basename(model)] = latency
        logging.result(f'[RESULT] predict latency for {os.path.basename(model)}: {latency} ms')
    
    return result


def get_nnmeter_ir_cli(args):
    """convert pb file or onnx file to nn-Meter IR graph according to the command line interface arguments
    """
    import json
    from nn_meter.utils.utils import NumpyEncoder
    if args.tensorflow:
        graph = model_file_to_graph(args.tensorflow, 'pb')
        filename = args.output if args.output else args.tensorflow.replace(".pb", "_pb_ir.json") 
    elif args.onnx:
        graph = model_file_to_graph(args.onnx, 'onnx')
        filename = args.output if args.output else args.onnx.replace(".onnx", "_onnx_ir.json") 
    else:
        logging.keyinfo('please run "nn-meter get_ir --help" to see guidance.')
        return
    
    if not str.endswith(filename, '.json'): filename += '.json'
    with open(filename, "w+") as fp:
        json.dump(graph,
            fp,
            indent=4,
            skipkeys=True,
            sort_keys=True,
            cls=NumpyEncoder,
        )
    
    logging.result(f'The nn-meter ir graph has been saved. Saved path: {os.path.abspath(filename)}')



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .prior_distribution_sampler import *
from .finegrained_sampler import *


class BaseConfigSampler:

    def prior_config_sampling(self, sample_num):
        ''' utilize the prior data to define the configuration sampling from the prior distribution.
        '''
        pass

    def finegrained_config_sampling(self, configs, sample_num):
        ''' for data in `configs`, perform fine-grained data sampling to generate random data around the large error data.
        '''
        pass


class ConvSampler(BaseConfigSampler):

    def prior_config_sampling(self, sample_num):
        return sampling_conv(sample_num)

    def finegrained_config_sampling(self, configs, sample_num):
        return finegrained_sampling_conv(configs, sample_num)


class DwConvSampler(BaseConfigSampler):

    def prior_config_sampling(self, sample_num):
        return sampling_dwconv(sample_num)

    def finegrained_config_sampling(self, configs, sample_num):
        return finegrained_sampling_dwconv(configs, sample_num)


class PoolingSampler(BaseConfigSampler):

    def prior_config_sampling(self, sample_num):
        return sampling_pooling(sample_num)

    def finegrained_config_sampling(self, configs, sample_num):
        return finegrained_sampling_pooling(configs, sample_num)


class FCSampler(BaseConfigSampler):

    def prior_config_sampling(self, sample_num):
        # half samples have fixed cout as 1000, other samples have random cout
        return sampling_fc(int(sample_num * 0.5), fix_cout = 1000) + sampling_fc(int(sample_num * 0.5), fix_cout = False)

    def finegrained_config_sampling(self, configs, sample_num):
        return finegrained_sampling_fc(configs, sample_num)


class ConcatSampler(BaseConfigSampler):

    def prior_config_sampling(self, sample_num):
        return sampling_concats(sample_num)

    def finegrained_config_sampling(self, configs, sample_num):
        return finegrained_sampling_concats(configs, sample_num)


class CinEvenSampler(BaseConfigSampler):

    def prior_config_sampling(self, sample_num):
        return sampling_hw_cin_even(sample_num)

    def finegrained_config_sampling(self, configs, sample_num):
        return finegrained_sampling_hw_cin_even(configs, sample_num)


class GlobalAvgPoolSampler(BaseConfigSampler):

    def prior_config_sampling(self, sample_num):
        cfgs = sampling_hw_cin(sample_num)
        new_hws = [3] * (sample_num // 2 + 1) + [7] * (sample_num // 2 + 1)
        new_hws = new_hws[:len(cfgs)]
        import random; random.shuffle(new_hws)
        for cfg, hw in zip(cfgs, new_hws): cfg["HW"] = hw
        return cfgs

    def finegrained_config_sampling(self, configs, sample_num):
        return finegrained_sampling_hw_cin(configs, sample_num)


class HwCinSampler(BaseConfigSampler):

    def prior_config_sampling(self, sample_num):
        return sampling_hw_cin(sample_num)
    
    def finegrained_config_sampling(self, configs, sample_num):
        return finegrained_sampling_hw_cin(configs, sample_num)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import json
from .fusion_lib import get_fusion_unit
from nn_meter.utils.graph_tool import ModelGraph


class RuleReader:
    rules_default = {
        "MON": 0,
        "FN": True,
    }

    multiop_blocks = ["se", "hswish", "channelshuffle", "gap"]

    def __init__(self, rule_file=None):
        self.rules = {}
        if rule_file:
            with open(rule_file, "r") as fp:
                self.rules = json.load(fp)
        self._extract_fusible()
        self._parse_multiop_block()

    def is_fusible(self, node_type, outnode_type):
        return (node_type, outnode_type) in self.fusible

    def query_rule(self, rule):
        if rule not in self.rules or self.rules[rule]["obey"] is None:
            return self.rules_default[rule]
        else:
            return self.rules[rule]["obey"]

    def _extract_fusible(self):
        def get_name(i):
            return f"{ops[i]}_{i}"

        self.fusible = []
        self.fusion_units = {}
        for name, rule in self.rules.items():
            if rule["obey"] and name.startswith("BF"):
                ops = name.split("_")[1:]
                if len(ops) == 2:
                    self.fusible.append((ops[0], ops[1]))
                elif len(ops) > 2:
                    fusion_unit = {}
                    for i in range(0, len(ops)):
                        fusion_unit[get_name(i)] = {
                            "attr": {
                                "type": ops[i],
                                "attr": {},
                            },
                            "inbounds": [get_name(i - 1)] if i > 0 else [],
                            "outbounds": [get_name(i + 1)] if i < len(ops) - 1 else [],
                        }
                    self.fusion_units["-".join(ops)] = [ModelGraph(graph=fusion_unit)]

    def _parse_multiop_block(self):
        for block in self.multiop_blocks:
            self.fusion_units[block] = get_fusion_unit(block)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import subprocess


def patch_frozenpb(graph_path, interpreter_path):
    """ 
    Patch a frozen pb file to make it compatible with Movidius VPU and then return the path to the patched pb file.
    @params:

    graph_path: Path to the frozen pb file.

    interpreter_path:  the path of python interpreter
    """
    scripts_dir = os.path.abspath(os.path.dirname(__file__))
    subprocess.run(
        f'{interpreter_path} {os.path.join(scripts_dir, "frozenpb_patcher.py")} {graph_path}',
        shell=True
    )
    return os.path.splitext(graph_path)[0] + '_patched.pb'



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import numpy as np

from .frozenpb_parser import FrozenPbParser
from .shape_inference import ShapeInference
from .shape_fetcher import ShapeFetcher
from nn_meter.utils.graph_tool import ModelGraph

class FrozenPbConverter:
    def __init__(self, file_name):
        self.model_graph = ModelGraph()

        # Parse pb to graph
        parser = FrozenPbParser(file_name)
        parser.parse_graph(self.model_graph)
        dynamic_fetcher = ShapeFetcher(parser.graph)

        # Change split to more firendly scheme
        parser.fix_split_naming(self.model_graph)

        # Get the static shape
        ShapeInference(self.model_graph, dynamic_fetcher)

        # Strip constant and indentity nodes
        parser.strip_useless_nodes(self.model_graph)

    def get_flatten_graph(self):
        def np_encoder(d):
            for k, v in d.items():
                if isinstance(v, dict):
                    np_encoder(v)
                else:
                    if isinstance(v, np.ndarray):
                        d[k] = v.tolist()
                    if isinstance(v, (bytes, bytearray)):
                        d[k] = v.decode("utf-8")

        np_encoder(self.model_graph.get_graph())
        return self.model_graph.get_graph()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import yaml
import logging
logging = logging.getLogger("nn-Meter")


__user_config_folder__ = os.path.expanduser('~/.nn_meter/config')
__registry_cfg_filename__ = 'registry.yaml'

def list_backends_cli():
    from nn_meter.builder.backends import list_backends
    backends = list_backends()
    logging.keyinfo("Supported backends: ('*' indicates customized backends)")
    for name in backends:
        logging.result(f"[Backend] {name}")


def list_operators_cli():
    from nn_meter.builder.backend_meta.fusion_rule_tester.utils import list_operators
    operators = list_operators()
    logging.keyinfo("Supported operators: ('*' indicates customized operators)")
    for name in operators:
        logging.result(f"[Operator] {name}")


def list_kernels_cli():
    from nn_meter.builder.kernel_predictor_builder.data_sampler.utils import list_kernels
    kernels = list_kernels()
    logging.keyinfo("Supported kernels: ('*' indicates customized kernels)")
    for name in kernels:
        logging.result(f"[Kernel] {name}")


def list_special_testcases_cli():
    from nn_meter.builder.backend_meta.fusion_rule_tester.generate_testcase import list_testcases
    testcases = list_testcases()
    logging.keyinfo("Supported testcases: ('*' indicates customized test cases)")
    for name in testcases:
        logging.result(f"[TestCase] {name}")


def create_workspace_cli(args):
    """create a workspace folder and copy the corresponding config file to the workspace
    """
    from nn_meter.builder.config_manager import copy_to_workspace
    
    if args.tflite_workspace:
        backend_type = "tflite"
        workspace_path = args.tflite_workspace
        copy_to_workspace(backend_type, workspace_path)
    elif args.openvino_workspace:
        backend_type = "openvino"
        workspace_path = args.openvino_workspace
        # create openvino_env
        openvino_env = os.path.join(workspace_path, 'openvino_env')
        os.system(f"virtualenv {openvino_env}")
        os.system("source {openvino_env}/bin/activate")
        os.system("pip install -r docs/requirements/openvino_requirements.txt")
        os.system("deactivate")
        copy_to_workspace(backend_type, workspace_path)
    elif args.customized_workspace:
        backend_type = "customized"
        backend_name = args.backend
        workspace_path = args.customized_workspace
        if os.path.isfile(os.path.join(__user_config_folder__, __registry_cfg_filename__)):
            with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'r') as fp:
                registry_modules = yaml.load(fp, yaml.FullLoader)
            try:
                backend_config = registry_modules["backends"][backend_name]["defaultConfigFile"]
                copy_to_workspace(backend_type, workspace_path, backend_config)
            except:
                raise ValueError(f"Create workspace failed. Please check the backend registration information.")
        else:
            raise ValueError(f"Create workspace failed. Please check the backend registration information.")
    else:
        logging.keyinfo('please run "nn-meter create --help" to see guidance.')
        return

    logging.keyinfo(f"Workspace {os.path.abspath(workspace_path)} for {backend_type} platform has been created. " \
        f"Users could edit experiment config in {os.path.join(os.path.abspath(workspace_path), 'configs/')}.")


def test_backend_connection_cli(args):
    from nn_meter.builder import builder_config
    from nn_meter.builder.backends import connect_backend
    if args.workspace and args.backend:
        builder_config.init(args.workspace)
        backend = connect_backend(args.backend)
        backend.test_connection()
    else:
        logging.keyinfo('please run "nn-meter connect --help" to see guidance.')



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import shutil
import logging
from ..interface import BaseBackend
from nn_meter.utils.path import get_filename_without_ext
logging = logging.getLogger("nn-Meter")


class TFLiteBackend(BaseBackend):
    parser_class = None
    profiler_class = None

    def update_configs(self):
        """update the config parameters for TFLite platform
        """
        super().update_configs()
        self.profiler_kwargs.update({
            'dst_graph_path': self.configs['REMOTE_MODEL_DIR'],
            'benchmark_model_path': self.configs['BENCHMARK_MODEL_PATH'],
            'serial': self.configs['DEVICE_SERIAL'],
            'dst_kernel_path': self.configs['KERNEL_PATH']
        })

    def convert_model(self, model_path, save_path, input_shape=None):
        """convert the Keras model instance to ``.tflite`` and return model path
        """
        import tensorflow as tf
        model_name = get_filename_without_ext(model_path)
        model = tf.keras.models.load_model(model_path)
        converter = tf.lite.TFLiteConverter.from_keras_model(model)
        tflite_model = converter.convert()
        converted_model = os.path.join(save_path, model_name + '.tflite')
        open(converted_model, 'wb').write(tflite_model)
        shutil.rmtree(model_path)
        return converted_model

    def test_connection(self):
        """check the status of backend interface connection, ideally including open/close/check_healthy...
        """
        from ppadb.client import Client as AdbClient
        client = AdbClient(host="127.0.0.1", port=5037)
        if self.configs['DEVICE_SERIAL']:
            device = client.device(self.configs['DEVICE_SERIAL'])
        else:
            device = client.devices()[0]
        logging.keyinfo(device.shell("echo hello backend !"))



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from setuptools import setup, find_packages


setup(
    name='nn-meter',
    version='2.0',
    description='nn-Meter is a novel and efficient system to accurately predict the inference latency of DNN models on diverse edge devices.',
    long_description = open('README.md', encoding='utf-8').read(),
    long_description_content_type = 'text/markdown',
    author='Microsoft nn-Meter Team',
    author_email='nn-meter-dev@microsoft.com',
    url='https://github.com/microsoft/nn-Meter',
    project_urls={
        'Data of models': 'https://github.com/microsoft/nn-Meter/releases/tag/v1.0-data',
    },
    license = 'MIT',
    classifiers = [
            'License :: OSI Approved :: MIT License',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: Microsoft :: Windows :: Windows 10',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
    packages=find_packages(),
    package_data={
        'nn_meter': [
            'configs/*.yaml',
            'configs/builder/backends/*.yaml',
            'configs/builder/fusion_rule_tester/*.yaml',
            'configs/builder/predictor_builder/*.yaml',
            'builder/kernel_predictor_builder/data_sampler/prior_config_lib/*.csv',
            'kernel_detector/fusion_lib/*.json'],
    },
    entry_points={
        'console_scripts': ['nn-meter=nn_meter.utils.nn_meter_cli.interface:nn_meter_cli'],
    },
    install_requires=[
        'numpy', 'pandas', 'tqdm', 'networkx', 'requests', 'protobuf', 'PyYAML', 'scikit_learn', 'packaging', 'jsonlines'
    ],
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

class BaseOperator:
    def __init__(self, input_shape = None, config = None):
        ''' base class for operator.
        
        @params
        
        - input_shape: defines the dimension of one model input shape without batch size. Generally, when the input shape is 3D, `input_shape`
            should be `[config["HW"], config["HW"], config["CIN"]]` (for tensorflow model), or `config["CIN"], [config["HW"], config["HW"]]` (for
            torch model), and when the input shape is 1D, `input_shape` should be`[config["CIN"]]`.

        - config: a dict containing all configurations.
        '''
        self.input_shape = input_shape
        self.config = config

    def get_model(self):
        pass

    def get_output_shape(self):
        return self.input_shape

    def get_is_two_inputs(self):
        return False

    def test_operator():
        ''' for users to test the model when registration. Do not need to override by users.
        '''
        pass


class BaseBlock:
    def __init__(self, config, batch_size = 1):
        ''' base class for kernel block.
        
        @params
        
        - input_shape: defines the dimension of one model input shape without batch size. Generally, when the input shape is 3D, `input_shape`
            should be `[config["HW"], config["HW"], config["CIN"]]` (for tensorflow model), or `config["CIN"], [config["HW"], config["HW"]]` (for
            torch model), and when the input shape is 1D, `input_shape` should be`[config["CIN"]]`.

        - input_tensor_shape: a list defining all model inputs. In basic situation, `input_tensor_shape` should be `[input_shape]` if the kernel
            only has one input. If the kernel has more than one input, such as `addrelu` kernel, `input_tensor_shape` is `[input_shape, input_shape]`.

        - batch_size: the required batch size of the input data
        '''
        self.config = config
        self.input_shape = None
        self.input_tensor_shape = None
        self.batch_size = batch_size

    def get_model(self):
        ''' the implementation of the kernel model and return a instance of `tensorflow.keras.Model` or `torch.nn.Module` of the kernel.
        '''
        pass

    def save_model(self, save_path):
        pass

    def test_block(self):
        pass


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .prediction.utils import latency_metrics
from .nn_meter_predictor import nnMeterPredictor, list_latency_predictors, load_latency_predictor



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import random
import numpy as np
from .prior_config_lib.utils import *


def inverse_transform_sampling(data, n_bins = 40, n_samples = 1000):
    ''' calculate inversed cdf, for sampling by possibility
    '''
    import scipy.interpolate as interpolate
    hist, bin_edges = np.histogram(data, bins=n_bins, density=True)
    cum_values = np.zeros(bin_edges.shape)
    cum_values[1:] = np.cumsum(hist*np.diff(bin_edges))
    inv_cdf = interpolate.interp1d(cum_values, bin_edges)
    r = np.random.rand(n_samples)
    data = inv_cdf(r)
    ndata = [int(x) for x in data]
    return ndata


def sample_based_on_distribution(data, count):
    ''' use data to calculate a inversed cdf, and sample `count` data from such distribution
    '''
    return inverse_transform_sampling(data, n_samples=count)


def data_validation(data, cdata):
    ''' convert sampled data to valid configuration, e.g.,: kernel size = 1, 3, 5, 7

    @params:
    data: the origin data value.
    cdata: valid configuration value.
    '''
    newlist = []
    for da in cdata:
        value = [abs(da - x) for x in data]
        newlist.append(value)

    newlist = list(np.asarray(newlist).T)    
    cda = [list(d).index(min(d)) for d in newlist]
    redata = [cdata[x] for x in cda]
    return redata


def sampling_conv(count):
    ''' 
    Sampling configs for conv kernels based on conv_zoo, which contains configuration values from existing model zoo for conv kernel. 
    The values are stored in prior_config_lib/conv.csv.
    Returned params include: (hw, cin, cout, kernel_size, strides)
    '''
    hws, cins, couts, kernel_sizes, strides = read_conv_zoo()
    new_cins = sample_based_on_distribution(cins, count)
    new_couts = sample_based_on_distribution(couts, count)

    # 70% of sampled data are from prior distribution
    count1 = int(count * 0.7)
    new_hws = sample_based_on_distribution(hws, count1)
    new_kernel_sizes = sample_based_on_distribution(kernel_sizes, count1)
    new_strides = sample_based_on_distribution(strides, count1)

    new_kernel_sizes = data_validation(new_kernel_sizes, [1, 3, 5, 7])
    new_strides = data_validation(new_strides, [1, 2, 4])
    new_hws = data_validation(new_hws, [1, 3, 7, 8, 13, 14, 27, 28, 32, 56, 112, 224])
    
    # since conv is the largest and most-challenging kernel, we add some frequently used configuration values
    new_hws.extend([112] * int((count - count1) * 0.2) + [56] * int((count - count1) * 0.4) + [28] * int((count - count1) * 0.4)) # frequent settings
    new_kernel_sizes.extend([5] * int((count - count1) * 0.4) + [7] * int((count - count1) * 0.6)) # frequent settings
    new_strides.extend([2] * int((count - count1) * 0.4) + [1] * int((count - count1) * 0.6)) # frequent settings
    random.shuffle(new_hws)
    random.shuffle(new_strides)
    random.shuffle(new_kernel_sizes)

    ncfgs = []
    for hw, cin, cout, kernel_size, stride in zip(new_hws, new_cins, new_couts, new_kernel_sizes, new_strides):
        c = {
            'HW': hw,
            'CIN': cin,
            'COUT': cout,
            'KERNEL_SIZE': kernel_size,
            'STRIDES': stride,
        }
        ncfgs.append(c)
    return ncfgs


def sampling_conv_random(count):
    ''' sampling configs for conv kernels based on random
    Returned params include: (hw, cin, cout, kernel_size, strides)
    '''
    hws = [1, 7, 8, 13, 14, 27, 28, 32, 56, 112, 224]
    kernel_sizes = [1, 3, 5, 7]
    strides = [1, 2, 4]
    
    cins = list(range(3, 2160))
    couts = list(range(16, 2048))
    new_hws = random.sample(hws * int(count / len(hws)) * 10, count)
    new_kernel_sizes = random.sample(kernel_sizes * int(count / len(kernel_sizes) * 10), count)
    new_strides = random.sample(strides * int(count / len(strides) * 10), count)
    new_cins = random.sample(cins * 10, count)
    new_couts = random.sample(couts * 18, count)
    random.shuffle(new_cins)
    random.shuffle(new_couts)

    ncfgs = []
    for hw, cin, cout, kernel_size, stride in zip(new_hws, new_cins, new_couts, new_kernel_sizes, new_strides):
        c = {
            'HW': hw,
            'CIN': cin,
            'COUT': cout,
            'KERNEL_SIZE': kernel_size,
            'STRIDES': stride,
        }
        ncfgs.append(c)
    return ncfgs


def sampling_dwconv(count):
    ''' 
    Sampling configs for dwconv kernels based on dwconv zoo, which contains configuration values from existing model zoo for dwconv kernel. 
    The values are stored in prior_config_lib/dwconv.csv.
    Returned params include: (hw, cin, kernel_size, strides)
    '''
    hws, cins, ks, strides = read_dwconv_zoo()
    new_cins = sample_based_on_distribution(cins, count)
   
    count1 = int(count * 0.8)
    new_hws = sample_based_on_distribution(hws,count1)
    new_kernel_sizes = sample_based_on_distribution(ks, count1)
    new_strides = sample_based_on_distribution(strides, count1)
    
    new_hws = data_validation(new_hws, [1, 3, 7, 14, 28, 56, 112, 224])
    new_kernel_sizes = data_validation(new_kernel_sizes, [1, 3, 5, 7])
    new_strides = data_validation(new_strides, [1, 2])
    
    new_hws.extend([112] * int((count - count1) * 0.4) + [56] * int((count - count1) * 0.4) + [28] * int((count - count1) * 0.2))  
    new_kernel_sizes.extend([5] * int((count - count1) * 0.4) + [7] * int((count - count1) * 0.6))
    new_strides.extend([2] * int((count - count1) * 0.5) + [1] * int((count - count1) * 0.5))
    random.shuffle(new_hws)
    random.shuffle(new_kernel_sizes)
    random.shuffle(new_strides)

    ncfgs = []
    for hw, cin, kernel_size, stride in zip(new_hws, new_cins, new_kernel_sizes, new_strides):
        c = {
            'HW': hw,
            'CIN': cin,
            'KERNEL_SIZE': kernel_size,
            'STRIDES': stride,
        }
        ncfgs.append(c)
    return ncfgs


def sampling_fc(count, fix_cout = 1000):
    '''
    Sampling configs for fc kernels based on fc zoo, which contains configuration values from existing model zoo for fc kernel. 
    The values are stored in prior_config_lib/fcs.csv.
    Returned params include: (cin, cout)
    '''
    cins, couts = read_fc_zoo()
    new_cins = sample_based_on_distribution(cins, count)
    if not fix_cout:
        new_couts = sample_based_on_distribution(couts, count)
    else:
        new_couts = [fix_cout] * count
    ncfgs = []
    for cin, cout in zip(new_cins, new_couts):
        c = {
            'CIN': cin,
            'COUT': cout,
        }
        ncfgs.append(c)
    return ncfgs


def sampling_pooling(count):
    '''
    Sampling configs for pooling kernels based on pooling zoo, which contains configuration values from existing model zoo for pooling kernel. 
    The values are stored in prior_config_lib/pooling.csv.
    Returned params include: (hw, cin, kernel_size, pool_strides)
    '''
    hws, cins, kernel_size, strides = read_pool_zoo()
    new_cins = sample_based_on_distribution(cins, count)
    new_hws = sample_based_on_distribution(hws, count)
    new_hws = data_validation(new_hws, [14, 28, 56, 112, 224])
    new_kernel_sizes = list(kernel_size) * (count // len(kernel_size) + 1)
    new_kernel_sizes = data_validation(new_kernel_sizes, [2, 3])
    random.shuffle(new_kernel_sizes)
    new_strides = list(strides) * (count // len(strides) + 1)
    new_strides = data_validation(new_strides, [1, 2])
    random.shuffle(new_strides)
    
    ncfgs = []
    for hw, cin, kernel_size, stride in zip(new_hws, new_cins, new_kernel_sizes, new_strides):
        c = {
            'HW': hw,
            'CIN': cin,
            'KERNEL_SIZE': kernel_size,
            'STRIDES': stride,
        }
        ncfgs.append(c)
    return ncfgs


def sampling_hw_cin(count):
    ''' sampling configs for kernels with hw and cin parameter
    Returned params include: (hw, cin)
    '''
    hws, cins, _, _, _ = read_conv_zoo()
    new_cins = sample_based_on_distribution(cins, count)
   
    count1 = int(count * 0.8)
    new_hws = sample_based_on_distribution(hws,count1)
    new_hws = data_validation(new_hws, [1, 3, 7, 14, 28, 56, 112, 224])
    new_hws.extend([112] * int((count - count1) * 0.4) + [56] * int((count - count1) * 0.4) + [28] * int((count - count1) * 0.2))
    random.shuffle(new_hws)

    ncfgs = []
    for hw, cin in zip(new_hws, new_cins):
        c = {
            'HW': hw,
            'CIN': cin,
        }
        ncfgs.append(c)
    return ncfgs


def sampling_hw_cin_even(count):
    ''' sampling configs for kernels with hw and cin (only even values) parameter, in case for split / se / channelshuffle
    Returned params include: (hw, cin)
    '''
    hws, cins, _, _, _ = read_conv_zoo()
    new_cins = sample_based_on_distribution(cins, count)
   
    count1 = int(count * 0.8)
    new_hws = sample_based_on_distribution(hws,count1)
    new_hws = data_validation(new_hws, [1, 3, 7, 14, 28, 56, 112, 224])
    new_hws.extend([112] * int((count - count1) * 0.4) + [56] * int((count - count1) * 0.4) + [28] * int((count - count1) * 0.2))
    random.shuffle(new_hws)

    ncfgs = []
    for hw, cin in zip(new_hws, new_cins):
        c = {
            'HW': hw,
            'CIN': cin + 1 if cin % 2 else cin,
        }
        ncfgs.append(c)
    return ncfgs


def sampling_concats(count):
    ''' sampling functions for concat kernel
    Returned params include: (hw, ns, cin1, cin2, cin3, cin4), ns are in [2, 4]
    '''
    hws, cins, _, _, _ = read_conv_zoo()
    new_hws = sample_based_on_distribution(hws, count)
    new_cins1 = sample_based_on_distribution(cins, count)
    new_cins2 = sample_based_on_distribution(cins, count)
    new_cins3 = sample_based_on_distribution(cins, count)
    new_cins4 = sample_based_on_distribution(cins, count)

    new_hws = data_validation(new_hws, [7, 14, 28, 56])  # current normals
    new_ns = [2] * (count - int(count * 0.4) - int(count * 0.2)) + [3] * int(count * 0.2) + [4] * int(count * 0.4)
    random.shuffle(new_ns)

    ncfgs = []
    for hw, n, cin1, cin2, cin3, cin4 in zip(new_hws, new_ns, new_cins1, new_cins2, new_cins3, new_cins4):
        cins = [cin1, cin2, cin3, cin4]
        onehot = [1] * n + [0] * (4 - n)
        onehot_cins = [x * y for x, y in zip(onehot, cins)]
        c = {
            'HW': hw,
            'CIN1': onehot_cins[0],
            'CIN2': onehot_cins[1],
            'CIN3': onehot_cins[2],
            'CIN4': onehot_cins[3]
        }
        ncfgs.append(c)
    return ncfgs



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import networkx as nx
from .generate_testcase import generate_testcases
from nn_meter.builder import builder_config

config = builder_config.get_module('ruletest')


class FusionRuleTester:
    def __init__(self):
        self._testcases = generate_testcases()

    def _build_dep_dag(self):
        dag = nx.DiGraph()

        for name, cls in self._testcases.items():
            dag.add_node(name)
            for dep in cls.deps:
                dag.add_edge(dep, name)

        self._dag = list(nx.topological_sort(dag))

    def generate(self):
        testcases = {}

        for name, cls in self._testcases.items():
            testcases[name] = cls(config).save_testcase()

        return testcases

    def analyze(self, profile_results):
        self._build_dep_dag()
        result = {}

        for name in self._dag:
            if name not in profile_results:
                continue

            result[name] = {}
            rule_cls = self._testcases[name]

            obey = True
            for dep, expect in rule_cls.deps.items():
                if result[dep]['obey'] != expect:
                    obey = False

            if obey:
                rule = rule_cls(config)
                rule.load_latency(profile_results[name])
                obey = rule.test()
                if config['DETAIL']:
                    latency = {key: str(value) for key, value in rule.latency.items()}
                    result[name]['latency'] = latency

            result[name]['obey'] = obey

        return result



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import copy
import json
import logging
from .utils import NumpyEncoder
logging = logging.getLogger("nn-Meter")


class ModelGraph:
    def __init__(self, filename=None, graph=None):
        if filename is not None:
            self.graph = json.load(open(filename, "r"))
        elif graph is not None:
            self.graph = copy.deepcopy(graph)
        else:
            self.graph = {}

    def node(self, name, inbound_nodes=None):
        self.graph[name] = {}
        if inbound_nodes is not None:
            self.graph[name]["inbounds"] = inbound_nodes
            for node in inbound_nodes:
                if node not in self.graph.keys():
                    self.graph[node] = {}
                if "outbounds" not in self.graph[node].keys():
                    self.graph[node]["outbounds"] = []
                self.graph[node]["outbounds"].append(name)

    def refresh(self):
        last_remove_nodes_cnt = -1
        while True:
            for name in self.graph.keys():
                self.graph[name]["outbounds"] = []

            for name in self.graph.keys():
                if "inbounds" in self.graph[name].keys():
                    for node in self.graph[name]["inbounds"]:
                        if node not in self.graph.keys():
                            while node in self.graph[name]["inbounds"]:
                                self.graph[name]["inbounds"].remove(node)
                        else:
                            if "outbounds" not in self.graph[node].keys():
                                self.graph[node]["outbounds"] = []

                            self.graph[node]["outbounds"].append(name)

            spare_nodes = []
            for name in self.graph.keys():
                if (
                    len(self.graph[name]["outbounds"]) == 0
                    and len(self.graph[name]["inbounds"]) == 0
                ):
                    spare_nodes.append(name)

            if last_remove_nodes_cnt == 0 and len(spare_nodes) == 0:
                break

            last_remove_nodes_cnt = len(spare_nodes)
            for removing_node_name in spare_nodes:
                del self.graph[removing_node_name]

    def get_graph(self):
        return self.graph

    def get_node_inbounds(self, name):
        if "inbounds" in self.graph[name]:
            return self.graph[name]["inbounds"]
        else:
            return []

    def get_node_outbounds(self, name):
        if "outbounds" in self.graph[name]:
            return self.graph[name]["outbounds"]
        else:
            return []

    def set_node_inbounds(self, name, inbounds):
        self.graph[name]["inbounds"] = inbounds

    def set_node_outbounds(self, name, outbounds):
        self.graph[name]["outbounds"] = outbounds

    def remove_node_inbounds(self, name, inbound):
        if inbound in self.graph[name]["inbounds"]:
            self.graph[name]["inbounds"].remove(inbound)

    def remove_node_outbounds(self, name, outbound):
        if outbound in self.graph[name]["outbounds"]:
            self.graph[name]["outbounds"].remove(outbound)

    def add_node_inbounds(self, name, inbound):
        self.graph[name]["inbounds"].append(inbound)

    def add_node_outbounds(self, name, outbound):
        self.graph[name]["outbounds"].append(outbound)

    def get_graph_head(self):
        self.heads = []
        for (key, value) in self.graph.items():
            if "inbounds" not in value.keys() or len(value["inbounds"]) == 0:
                self.heads.append(key)
        return self.heads

    def get_graph_tail(self):
        self.tails = []
        for (key, value) in self.graph.items():
            if "outbounds" not in value.keys() or len(value["outbounds"]) == 0:
                self.tails.append(key)
        return self.tails

    def add_node_attr(self, name, attr_key, attr_value):
        if name not in self.graph.keys():
            self.graph[name] = {}
        self.graph[name]["attr"]["attr"][attr_key] = attr_value

    def set_node_attr(self, name, attr):
        if name not in self.graph.keys():
            self.graph[name] = {}
        self.graph[name]["attr"] = attr

    def get_node_attr(self, name):
        if name in self.graph.keys():
            return self.graph[name]["attr"]
        else:
            return None

    def get_node_type(self, name):
        if name in self.graph.keys() and "attr" in self.graph[name].keys():
            return self.graph[name]["attr"]["type"]
        else:
            logging.info(name, self.graph[name])
            return None

    def get_root_node(self, subgraph):
        root = next(iter(subgraph))

        flag = True
        while flag:
            flag = False
            for inbound in self.get_node_inbounds(root):
                if inbound in subgraph:
                    flag = True
                    root = inbound
                    break

        return root

    def fuse(self, subgraph, type, name=None, attr=None, is_block=True):
        """
        subgraph: list of node name
        Nothing will be done if subgraph doesn't exist in self
        """
        for node in subgraph:
            if node not in self.graph:
                return False

        if name is None:
            name = ";".join(subgraph)

        if attr is None:
            root_node = self.get_root_node(subgraph)
            attr = self.get_node_attr(root_node)
        attr["type"] = type
        if is_block:
            attr["attr"]["primitive_nodes"] = list(subgraph)

        self.graph[name] = {
            "attr": attr,
            "inbounds": [],
            "outbounds": [],
        }

        for node in subgraph:
            for inbound in self.get_node_inbounds(node):
                if inbound not in subgraph:
                    if inbound not in self.get_node_inbounds(name):
                        self.add_node_inbounds(name, inbound)
                    self.remove_node_outbounds(inbound, node)
                    if name not in self.get_node_outbounds(inbound):
                        self.add_node_outbounds(inbound, name)
            for outbound in self.get_node_outbounds(node):
                if outbound not in subgraph:
                    if outbound not in self.get_node_outbounds(name):
                        self.add_node_outbounds(name, outbound)
                    self.remove_node_inbounds(outbound, node)
                    if name not in self.get_node_inbounds(outbound):
                        self.add_node_inbounds(outbound, name)

        for node in subgraph:
            del self.graph[node]

        return True

    def plot_graphs(self, comment="Network Graph View"):
        from graphviz import Digraph

        dot = Digraph(comment=comment)
        for (key, value) in self.graph.items():
            dot.node(key, key)
            if "inbounds" in value.keys():
                for node in value["inbounds"]:
                    dot.edge(
                        node,
                        key,
                        label=", ".join(str(x) for x in value["attr"]["output_shape"]),
                    )
        dot.render("graph.gv", view=False)

    def plot_networkx_graph(self):
        import matplotlib.pyplot as plt
        import networkx as nx

        plt.subplot(121)
        nx.draw(self.get_networkx_graph(), with_labels=True, font_weight="bold")
        plt.show()

    def get_networkx_graph(self):
        import networkx as nx

        G = nx.MultiDiGraph()
        for (key, value) in self.graph.items():
            G.add_node(key, type=value["attr"]["type"], **value["attr"]["attr"])
            if "inbounds" in value.keys():
                for node in value["inbounds"]:
                    G.add_edge(node, key)
        self.graphx = G
        return G

    def match_isomorph_vf2(self):
        pass

    def find_subgraphs(self, sub_graph, match_func):
        from networkx.algorithms import isomorphism as iso

        GM = iso.MultiDiGraphMatcher(
            self.get_networkx_graph(),
            sub_graph.get_networkx_graph(),
            node_match=match_func,
        )

        matches = []
        for match in GM.subgraph_isomorphisms_iter():
            matches.append(
                {
                    key: value
                    for key, value in match.items()
                    if sub_graph.get_node_type(value) != "dummy"
                }
            )
        return matches

    def find_weight_roots(self, layer_name):
        weight_roots = []
        weights_nodes = []
        for inbound in self.graph[layer_name]["inbounds"]:
            if (
                self.graph[inbound]["attr"]["type"] == "Identity"
                and len(self.graph[inbound]["inbounds"]) == 1
            ):
                if (
                    self.graph[self.graph[inbound]["inbounds"][0]]["attr"]["type"]
                    == "Const"
                ):
                    weight_roots.append(inbound)
                    weights_nodes.append(inbound)
                    weights_nodes.append(self.graph[inbound]["inbounds"][0])

            if (
                self.graph[inbound]["attr"]["type"] == "Const"
                and len(self.graph[inbound]["inbounds"]) == 0
            ):
                weight_roots.append(inbound)
                weights_nodes.append(inbound)

        return weight_roots, weights_nodes

    def get_subgraphs(self, sub_graph, match_func):
        import tensorflow as tf
        import copy

        fetched_subgraphs = self.find_subgraphs(sub_graph, match_func)
        tar_sub_graphs = []
        for sub_fetch_graph in fetched_subgraphs:
            tar_sub_graphs.append(tf.GraphDef())

            for op_entry in sub_fetch_graph.keys():
                # --- Repleace dummy op ---
                if (
                    sub_graph.get_graph()[sub_fetch_graph[op_entry]]["attr"]["type"]
                    == "dummy"
                ):
                    dummy_op = tar_sub_graphs[-1].node.add()
                    dummy_op.op = "Identity"
                    dummy_op.name = sub_fetch_graph[op_entry]
                    dummy_op.input.extend(
                        sub_graph.get_graph()[sub_fetch_graph[op_entry]]["inbounds"]
                    )
                    dummy_op.attr["T"].type = 1
                else:
                    # --- Fetch the main op ---
                    node = copy.deepcopy(self.graph[op_entry]["attr"]["node"])

                    node.name = sub_fetch_graph[op_entry]

                    del node.input[:]
                    node.input.extend(
                        sub_graph.get_graph()[sub_fetch_graph[op_entry]]["inbounds"]
                    )
                    # --- Fetch the constant op ---
                    roots, nodes = self.find_weight_roots(op_entry)
                    for weight_root in roots:
                        node.input.append(weight_root)

                    for weight_node in nodes:
                        tar_sub_graphs[-1].node.append(
                            self.graph[weight_node]["attr"]["node"]
                        )

                    tar_sub_graphs[-1].node.append(node)

            # tf.io.write_graph(tar_sub_graphs[-1], '', 'a.pb')
        return tar_sub_graphs

    def dump_json(self, filename):
        with open(filename, "w+") as fp:
            json.dump(
                self.graph,
                fp,
                indent=4,
                skipkeys=True,
                sort_keys=True,
                cls=NumpyEncoder,
            )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import networkx as nx
from .union_find import UF
from nn_meter.utils.graph_tool import ModelGraph


class FusionAwareGraph:
    def __init__(self, model_graph: ModelGraph):
        self._model_graph = model_graph
        self._dag = list(nx.topological_sort(model_graph.get_networkx_graph()))
        self._uf = UF(len(self._dag))

        reverse = {}
        for index, name in enumerate(self._dag):
            reverse[name] = index
        outbounds = []
        inbounds = []
        for index, name in enumerate(self._dag):
            outbounds.append(
                {reverse[outbound] for outbound in self._model_graph.get_node_outbounds(name)}
            )
            inbounds.append(
                {reverse[inbound] for inbound in self._model_graph.get_node_inbounds(name)}
            )

        self._outbounds = outbounds
        self._inbounds = inbounds
        self._ready = [not inbounds[i] for i in range(0, len(self))]
        self._types = [model_graph.get_node_type(name) for name in self._dag]

    @property
    def nodes(self):
        return self._dag

    def __len__(self):
        return len(self._dag)

    def __getitem__(self, key):
        return self._dag[key]

    def fuse(self, node, outnode, update=False):
        """
        node should be root, outnode should be an unfused single node
        """
        self._uf.union(node, outnode)
        if not update:
            self._outbounds[node] = self._outbounds[outnode]
        else:
            self._outbounds[node].update(self._outbounds[outnode])

    def mark_ready(self, node):
        self._ready[node] = True

    def is_ready(self, node):
        for inbound in self._inbounds[node]:
            if not self.is_ready[inbound]:
                return False
        return True

    def is_visited(self, node):
        return self._ready[node]

    def get_outbounds(self, node):
        return self._outbounds[node]

    def get_inbounds(self, node):
        return self._inbounds[node]

    def get_type(self, node):
        return self._types[node]

    def get_basicblocks(self):
        bbs = []

        for _ in range(0, len(self)):
            bbs.append([])

        for i in range(0, len(self)):
            root = self._uf.find(i)
            bbs[root].append(self[i])

        bbs = [bb for bb in bbs if bb]
        return bbs

    def find_root(self, node):
        return self[self._uf.find(node)]

    def is_fused(self, node):
        return self._uf.find(node) != node

    def is_connected(self, p, q):
        return self._uf.connected(p, q)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .rule_reader import RuleReader
from .utils.match_helper import MatchHelper
from .utils.fusion_aware_graph import FusionAwareGraph
from nn_meter.utils.graph_tool import ModelGraph


class RuleSplitter:
    def __init__(self, rule_reader: RuleReader):
        self.rule_reader = rule_reader

    def fuse_multiop_blocks(self, model_graph: ModelGraph):
        for type, blocks in self.rule_reader.fusion_units.items():
            for block in blocks:
                subgraphs = model_graph.find_subgraphs(block, MatchHelper.op_type_matcher)
                for subgraph in subgraphs:
                    model_graph.fuse(subgraph.keys(), type)

    def split(self, model_graph: ModelGraph):
        """
        Apply rules to graph
        """
        self.preprocess(model_graph)
        fusion_graph = FusionAwareGraph(model_graph)

        i = -1
        while i < len(fusion_graph) - 1:
            i += 1
            if fusion_graph.is_fused(i):
                continue
            fusion_graph.mark_ready(i)
            if not fusion_graph.get_outbounds(i):
                continue
            # MON
            mon = self.rule_reader.query_rule("MON")
            if mon == 0:  # can't fuse if having multiple out node
                if len(fusion_graph.get_outbounds(i)) > 1:
                    continue
            # FN: TODO: which one is the first node
            fused = False
            for j in fusion_graph.get_outbounds(i):
                if fusion_graph.is_fused(j):
                    continue
                outnode_type = fusion_graph.get_type(j)
                node_type = fusion_graph.get_type(i)
                if not self.rule_reader.is_fusible(node_type, outnode_type):
                    continue
                # fuse node
                if mon == 0:
                    fusion_graph.fuse(i, j)
                else:
                    fusion_graph.fuse(i, j, True)
                fusion_graph.mark_ready(j)
                fused = True
                if mon == 1:  # only fused to first outnode
                    break
            if fused:
                i -= 1

        self._fusion_graph = fusion_graph
        return fusion_graph.get_basicblocks()

    def preprocess(self, model_graph: ModelGraph):
        self.fuse_multiop_blocks(model_graph)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import sys
import yaml
import importlib


__BUILTIN_BACKENDS__ = {
    "tflite_cpu": {
        "class_module": "nn_meter.builder.backends.tflite",
        "class_name": "TFLiteCPUBackend"
    },
    "tflite_gpu": {
        "class_module": "nn_meter.builder.backends.tflite",
        "class_name": "TFLiteGPUBackend"
    },
    "openvino_vpu": {
        "class_module": "nn_meter.builder.backends.openvino",
        "class_name": "OpenVINOVPUBackend"
    },
    "debug_backend": {
        "class_module": "nn_meter.builder.backends.interface",
        "class_name": "DebugBackend"
    }
}

__user_config_folder__ = os.path.expanduser('~/.nn_meter/config')
__registry_cfg_filename__ = 'registry.yaml'
__REG_BACKENDS__ = {}
if os.path.isfile(os.path.join(__user_config_folder__, __registry_cfg_filename__)):
    with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'r') as fp:
        registry_modules = yaml.load(fp, yaml.FullLoader)
    if "backends" in registry_modules:
        __REG_BACKENDS__ = registry_modules["backends"]


class BaseBackend:
    """
    the base backend class to instantiate a backend instance. If users want to implement their own backend,
    the customized Backend should inherit this class.

    @params:

    profiler_class: a subclass inherit form `nn_meter.builder.backend.BaseProfiler` to specify the running command of
        the backend. A profiler contains commands to push the model to mobile device, run the model on the mobile device,
        get stdout from the mobile device, and related operations. In the implementation of a profiler, an interface of
        ``Profiler.profile()`` is required.
    
    parser_class: a subclass inherit form `nn_meter.builder.backend.BaseParser` to parse the profiled results.
        A parser parses the stdout from devices profiler and get required metrics. In the implementation of a parser, interface
        of `Parser.parse()` and property of `Parser.results()` are required.
    """
    profiler_class = None
    parser_class = None

    def __init__(self, configs):
        """ class initialization with required configs
        """
        self.configs = configs
        self.update_configs()
        if self.parser_class:
            self.parser = self.parser_class(**self.parser_kwargs)
        if self.profiler_class:
            self.profiler = self.profiler_class(**self.profiler_kwargs)

    def update_configs(self):
        """ update the config parameters for the backend
        """
        self.parser_kwargs = {}
        self.profiler_kwargs = {}
    
    def convert_model(self, model_path, save_path, input_shape=None):
        """ convert the Keras model instance to the type required by the backend inference.

        @params:
        
        model_path: the path of model waiting to profile
        
        save_path: folder to save the converted model
        
        input_shape: the shape of input tensor for inference, a random tensor according to the shape will be 
            generated and used
        """
        # convert model and save the converted model to path `converted_model`
        converted_model = model_path
        return converted_model

    def profile(self, converted_model, metrics = ['latency'], input_shape = None, **kwargs):
        """
        run the model on the backend, return required metrics of the running results. nn-Meter only support latency
        for metric by now. Users may provide other metrics in their customized backend.

        @params:

        converted_model: the model path in type of backend required
        
        metrics: a list of required metrics name. Defaults to ['latency']
        
        """
        return self.parser.parse(self.profiler.profile(converted_model, **kwargs)).results.get(metrics)

    def profile_model_file(self, model_path, save_path, input_shape = None, metrics = ['latency'], **kwargs):
        """ load model by model file path, convert model file, and run ``self.profile()``
        @params:

        model_path: the path of model waiting to profile
        
        save_path: folder to save the converted model
        
        input_shape: the shape of input tensor for inference, a random tensor according to the shape will be 
            generated and used
        """
        converted_model = self.convert_model(model_path, save_path, input_shape)
        res = self.profile(converted_model, metrics, input_shape, **kwargs)
        return res

    def test_connection(self):
        """ check the status of backend interface connection.
        """
        pass


class BaseProfiler:
    """
    Specify the profiling command of the backend. A profiler contains commands to push the model to mobile device, run the model 
    on the mobile device, get stdout from the mobile device, and related operations. 
    """
    def profile(self):
        """ Main steps of ``Profiler.profile()`` includes 1) push the model file to edge devices, 2) run models in required times
        and get back running results. Return the running results on edge device.
        """
        output = ''
        return output


class BaseParser:
    """
    Parse the profiled results. A parser parses the stdout from devices runner and get required metrics.
    """
    def parse(self, content):
        """ A string parser to parse profiled results value from the standard output of devices runner. This method should return the instance
        class itself.

        @params
        
        content: the standard output from device       
        """
        return self

    @property
    def results(self):
        """ warp the parsed results by ``ProfiledResults`` class from ``nn_meter.builder.backend_meta.utils`` and return the parsed results value.
        """
        pass

class DebugBackend(BaseBackend):
    """ For debug use when there is no backend available. All latency value are randomly generated.
    """
    
    def profile(self, converted_model, metrics = ['latency'], input_shape = None, **kwargs):
        import random
        from nn_meter.builder.backend_meta.utils import Latency, ProfiledResults
        latency = Latency(random.randrange(0, 10000) / 100, random.randrange(0, 1000) / 1000) 
        return ProfiledResults({'latency': latency}).get(metrics)

    def test_connection(self):
        """ check the status of backend interface connection.
        """
        import logging
        logging.info("hello backend !")


def connect_backend(backend_name):
    """ 
    Return the required backend class, and feed params to the backend. Supporting backend: tflite_cpu, tflite_gpu, openvino_vpu.
    
    Available backend and corresponding configs: 
    - For backend based on TFLite platform: {
        'REMOTE_MODEL_DIR': path to the folder (on mobile device) where temporary models will be copied to.
        'BENCHMARK_MODEL_PATH': path (on android device) where the binary file `benchmark_model` is deployed.
        'DEVICE_SERIAL': if there are multiple adb devices connected to your host, you need to provide the \\
                         corresponding serial id. Set to '' if there is only one device connected to your host.
        'KERNEL_PATH': path (on mobile device) where the kernel implementations will be dumped.
    }
    - For backend based on OpenVINO platform: {
        'OPENVINO_ENV': path to openvino virtualenv (./docs/requirements/openvino_requirements.txt is provided)
        'OPTIMIZER_PATH': path to openvino optimizer
        'OPENVINO_RUNTIME_DIR': directory to openvino runtime
        'DEVICE_SERIAL': serial id of the device
        'DATA_TYPE': data type of the model (e.g., fp16, fp32)
    }
    
    The config can be declared and modified after create a workspace. Users could follow guidance from ./docs/builder/backend.md
    
    @params:
    backend_name: name of backend (subclass instance of `BaseBackend`). 
    """
    if backend_name in __REG_BACKENDS__:
        backend_info = __REG_BACKENDS__[backend_name]
        sys.path.append(backend_info["package_location"])
    elif backend_name in __BUILTIN_BACKENDS__:
        backend_info = __BUILTIN_BACKENDS__[backend_name]
    else:
        raise ValueError(f"Unsupported backend name: {backend_name}. Please register the backend first.")

    module = backend_info["class_module"]
    name = backend_info["class_name"]
    backend_module = importlib.import_module(module)   
    backend_cls = getattr(backend_module, name)

    # load configs from workspace
    from nn_meter.builder import builder_config
    configs = builder_config.get_module('backend')
    return backend_cls(configs)


def list_backends():
    """ list all backends supported by nn-Meter, including builtin backends and registered backends
    """
    return list(__BUILTIN_BACKENDS__.keys()) + ["* " + item for item in list(__REG_BACKENDS__.keys())]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import json
import logging
from ..utils import read_profiled_results
from nn_meter.builder.utils import merge_info
from nn_meter.builder.backend_meta.utils import Latency


class BaseTestCase:
    name = ''
    cases = None
    true_case = ''
    deps = {}
    input_shape = None
    implement = None

    def __init__(self, config, **kwargs):
        self._kwargs = kwargs
        self.latency = {}
        self.config = config
        self.load_config()

    def generate_testcase(self):
        testcase = {}
        model, shapes = self._model_block()
        testcase['block'] = {
            'model': model,
            'shapes': shapes,
        }

        for _, ops in self.cases.items():
            for op in ops:
                try:
                    model, shapes = getattr(self, '_model_' + op)()
                    testcase[op] = {
                        'model': model,
                        'shapes': shapes
                    }
                except:
                    from .utils import generate_single_model
                    model, shapes = generate_single_model(op, self.input_shape, self.config, self.implement)
                    testcase[op] = {
                        'model': model,
                        'shapes': shapes
                    }
        return testcase

    def save_testcase(self):
        from .utils import save_model
        testcase = self.generate_testcase()

        for op, model in testcase.items():
            model_path = os.path.join(self.workspace_path, self.name + '_' + op)
            model_path = save_model(model, model_path, self.implement)
            testcase[op]['model'] = model_path

        return testcase

    def load_latency(self, testcase):
        self.latency['block'] = Latency(testcase['block']['latency'])

        for case, ops in self.cases.items():
            latency_sum = 0
            for op in ops:
                if op not in self.latency:
                    self.latency[op] = Latency(testcase[op]['latency'])
                latency_sum += self.latency[op]
            self.latency[case] = latency_sum

    def test(self):
        true_case_lat_diff = abs(self.latency[self.true_case].avg - self.latency['block'].avg)

        for case, _ in self.cases.items():
            if case != self.true_case and true_case_lat_diff > abs(self.latency[case].avg - self.latency['block'].avg):
                return case

        return self.true_case

    def load_config(self):
        config = self.config
        if not self.input_shape:
            self.input_shape = [config['HW'], config['HW'], config['CIN']]
        self.kernel_size = config['KERNEL_SIZE']
        self.cout = config['COUT']
        self.padding = config['PADDING']
        self.workspace_path = os.path.join(config['WORKSPACE'], 'testcases')
        os.makedirs(self.workspace_path, exist_ok=True)

    def _model_block(self):
        pass


def generate_testcases():
    """generate testcases and save the testcase models and testcase json file in the workspace
    Users could edit the configurations of testcases in <workspace-path>/configs/ruletest_config.yaml.
    The config will take effect after the the config file is saved and closed.
    """
    from nn_meter.builder import builder_config
    config = builder_config.get_module('ruletest')

    from .test_fusion_rule import FusionRuleTester
    tester = FusionRuleTester()
    testcases = tester.generate()

    # save information to json file
    workspace_path = config['WORKSPACE']
    info_save_path = os.path.join(workspace_path, "results", "origin_testcases.json")
    new_testcases = merge_info(new_info=testcases, info_save_path=info_save_path)
    os.makedirs(os.path.dirname(info_save_path), exist_ok=True)
    with open(info_save_path, 'w') as fp:
        json.dump(new_testcases, fp, indent=4)
    logging.keyinfo(f"Save the original testcases information to {info_save_path}")
    return testcases


def detect_fusion_rule(profiled_testcases):
    """ detect fusion rule by testcases latency value
    @params:

    testcases: the Dict of testcases or the path of the testcase json file
    """
    if isinstance(profiled_testcases, str):
        with open(profiled_testcases, 'r') as fp:
            profiled_testcases = read_profiled_results(json.load(fp))

    from .test_fusion_rule import FusionRuleTester
    tester = FusionRuleTester()
    result = tester.analyze(profiled_testcases)

    # save information to json file
    from nn_meter.builder import builder_config
    config = builder_config.get_module('ruletest')
    workspace_path = config['WORKSPACE']
    info_save_path = os.path.join(workspace_path, "results", "detected_fusion_rule.json")
    new_result = merge_info(new_info=result, info_save_path=info_save_path)
    os.makedirs(os.path.dirname(info_save_path), exist_ok=True)
    with open(info_save_path, 'w') as fp:
        json.dump(new_result, fp, indent=4)
    logging.keyinfo(f"Save the detected fusion rule information to {info_save_path}")
    return result



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import sys
import yaml
import json
import logging
import importlib
from nn_meter.builder.backend_meta.utils import read_profiled_results
logging = logging.getLogger("nn-Meter")


feature_for_kernel = {
    # conv
    "conv-bn-relu":         ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "conv-bn-relu6":        ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "conv-bn":              ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "conv-relu":            ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "conv-relu6":           ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "conv-hswish":          ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "conv-block":           ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "conv-bn-hswish":       ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    # dwconv ("COUT" will always be the same as "CIN")
    "dwconv-bn":            ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "dwconv-relu":          ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "dwconv-relu6":         ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "dwconv-bn-relu":       ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "dwconv-bn-relu6":      ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "dwconv-block":         ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    "dwconv-bn-hswish":     ["HW", "CIN", "COUT", "KERNEL_SIZE", "STRIDES"],
    # pooling ("COUT" will always be the same as "CIN")
    "maxpool":              ["HW", "CIN", "COUT", "KERNEL_SIZE", "POOL_STRIDES"],
    "avgpool":              ["HW", "CIN", "COUT", "KERNEL_SIZE", "POOL_STRIDES"],
    # others
    "fc":                   ["CIN", "COUT"],
    "concat":               ["HW", "CIN1", "CIN2", "CIN3", "CIN4"],
    "split":                ["HW", "CIN"],
    "channelshuffle":       ["HW", "CIN"],
    "se":                   ["HW", "CIN"],
    "global-avgpool":       ["HW", "CIN"],
    "bnrelu":               ["HW", "CIN"],
    "bn":                   ["HW", "CIN"],
    "hswish":               ["HW", "CIN"],
    "relu":                 ["HW", "CIN"],
    # In "addrelu" block and "add" block, the second feature "CIN" will always be the same as
    # the third feature
    "addrelu":              ["HW", "CIN", "CIN"],
    "add":                  ["HW", "CIN", "CIN"], 
}

__user_config_folder__ = os.path.expanduser('~/.nn_meter/config')
__registry_cfg_filename__ = 'registry.yaml'
__REG_KERNELS__ = {}
if os.path.isfile(os.path.join(__user_config_folder__, __registry_cfg_filename__)):
    with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'r') as fp:
        registry_modules = yaml.load(fp, yaml.FullLoader)
    if "kernels" in registry_modules:
        __REG_KERNELS__ = registry_modules["kernels"]


class BaseFeatureParser:
    def __init__(self, kernel_type):
        self.kernel_type = kernel_type
        self.needed_config = feature_for_kernel[kernel_type]

    def get_feature_by_config(self, config_dict):
        if "COUT" not in config_dict and "COUT" in self.needed_config:
            config_dict["COUT"] = config_dict["CIN"]
        feature = [config_dict[data] for data in self.needed_config]
        return feature

    def get_config_by_feature(self, feature):
        assert len(self.needed_config) == len(feature)
        config = {k: v for k, v in zip(self.needed_config, feature)}
        return config


class FlopsParamParser(BaseFeatureParser):
    def get_feature_by_config(self, config_dict):
        if "COUT" not in config_dict and "COUT" in self.needed_config:
            config_dict["COUT"] = config_dict["CIN"]
        feature = [config_dict[data] for data in self.needed_config]
        from .utils import get_flops_params
        flop, param = get_flops_params(self.kernel_type, config_dict)
        flop /= 2e6
        param /= 1e6
        feature.extend([flop, param])
        return feature

    def get_config_by_feature(self, feature):
        # remove flops and params num feature from feature vector
        feature = feature[:-2]
        assert len(self.needed_config) == len(feature)
        config = {k: v for k, v in zip(self.needed_config, feature)}
        return config


def get_feature_parser(kernel_type):
    if kernel_type in __REG_KERNELS__:
        kernel_info = __REG_KERNELS__[kernel_type]
        sys.path.append(kernel_info["package_location"])
        parser_name = kernel_info["parser_name"]
        parser_module = importlib.import_module(kernel_info["parser_module"])
        return getattr(parser_module, parser_name)(kernel_type)
    elif kernel_type in feature_for_kernel:
        if "conv" in kernel_type or "dwconv" in kernel_type or "fc" in kernel_type:
            return FlopsParamParser(kernel_type)
        else:
            return BaseFeatureParser(kernel_type)


def get_data_by_profiled_results(kernel_type, feature_parser, cfgs_path, labs_path = None,
                                 save_path = None, predict_label = "latency"):
    ''' return (features, latency)
    kernel_type (str): type of kernel
    
    feature_parser (subclass instance of BaseFeatureParser) the parser containing the feature parsing script

    cfgs_path (str or dict): path of config information dict, or dict of "origin_kernels.json", such as
        {
            "conv-bn-relu": {
                "id_0": {
                    "model": "...",
                    "shapes": [[14, 14, 98]],
                    "config": {
                        "HW": 14,
                        "CIN": 98,
                        "COUT": 120,
                        "KERNEL_SIZE": 3,
                        "STRIDES": 1
                    },
                }
            }
        }

    labs_path (str or dict): path of profiled label information dict, or dict of "profiled_results", such as
        {
            "conv-bn-relu": {
                "id_0": {
                    "latency": "42.001 +- 1.0"
                }
            }
        }
        if labs_path == None, it means latency (or other label) information are also included in cfgs_path.

    save_path (str): the path to save the feature and latency information, including file name. If save_path is None, the data will not be saved.

    predict_label (str): the predicting label to build kernel predictor
    '''
    if labs_path == None:
        if type(cfgs_path) == tuple:
            cfgs_path, labs_path = cfgs_path
        else:
            labs_path = cfgs_path
    if isinstance(cfgs_path, str):
        with open(cfgs_path, 'r') as fp:
            cfgs_dict = json.load(fp)[kernel_type]
    else:
        cfgs_dict = cfgs_path[kernel_type] if kernel_type in cfgs_path else cfgs_path
    if isinstance(labs_path, str):
        with open(labs_path, 'r') as fp:
            labs_dict = read_profiled_results(json.load(fp))[kernel_type]
    else:
        labs_dict = labs_path[kernel_type] if kernel_type in labs_path else labs_path

    paths, features, labs = [], [], []
    for id in labs_dict.keys():
        try:
            path = cfgs_dict[id]["model"]
            configs = cfgs_dict[id]["config"]
            feature = feature_parser.get_feature_by_config(configs)
            if predict_label == "latency":
                label = labs_dict[id]["latency"].avg
            else:
                label = labs_dict[id][predict_label]
            if label != 0.0:
                paths.append(os.path.basename(path))
                features.append(feature)
                labs.append(label)
        except:
            pass

    # save features and latency information to `save_path`
    if save_path:
       import pandas as pd
       cols = feature_parser.needed_config[:]
       if len(features[0]) - len(feature_parser.needed_config) > 0: # there are extra features beyond needed config
           cols += [f'feature_{i}' for i in range(len(features[0]) - len(feature_parser.needed_config))]
       data_df = pd.DataFrame(features, columns=cols)
       data_df = pd.concat([pd.DataFrame(paths, columns=["model_path"]), data_df], axis=1)
       data_df[predict_label] = labs
       data_df.to_csv(save_path, index=False)
       logging.info(f'Saved the feature table of all data for {kernel_type} in path {save_path}.')

    return (features, labs)
